import re
import typing

from .._types import RE_FLAGS_ALIAS
from .._types import RE_PATTERN_ALIAS
from ..descriptors import ValidatesInstanceOf
from ._handler import Handler


class RegexHandler(Handler):
    """
    Regular expression handler.
    """

    actual: typing.Any = ValidatesInstanceOf(str, re.Pattern)

    def __init__(self, actual: typing.Any) -> None:
        super().__init__(actual)

    def match(self, pattern: RE_PATTERN_ALIAS, flags: RE_FLAGS_ALIAS = 0) -> None:
        """Matches the beginning of a string"""
        if not re.match(pattern, self.actual, flags):
            raise AssertionError(f"{self.actual} did not begin with pattern: {pattern=}")

    def does_not_match(self, pattern: RE_PATTERN_ALIAS, flags: RE_FLAGS_ALIAS = 0) -> None:
        try:
            self.match(pattern, flags)
            raise AssertionError(f"{self.actual} was a match with pattern: {pattern=}")
        except AssertionError:
            pass

    def search(self, pattern: RE_PATTERN_ALIAS, flags: RE_FLAGS_ALIAS) -> None:
        if re.search(pattern, self.actual, flags) is None:
            raise AssertionError(f"{self.actual} did not contain any matches for: {pattern=}")

    def fullmatch(self, pattern: RE_PATTERN_ALIAS, flags: RE_FLAGS_ALIAS) -> None:
        if re.fullmatch(pattern, self.actual, flags) is None:
            raise AssertionError(f"{self.actual} was not matched entirely by: {pattern=}")

    def findall(self, pattern: RE_PATTERN_ALIAS, count: int, flags: RE_FLAGS_ALIAS) -> None:
        matches = re.findall(pattern, self.actual, flags)
        if len(matches) != count:
            raise AssertionError(
                f"{self.actual} had: {len(matches)} non overlapping occurrences for pattern: {pattern}, not: {count}"
            )



"""
Asserto exception hierarchy.
    * ActualTypeError
"""
import typing

from .handlers import Handler


class ActualTypeError(TypeError):
    """Raised when type error problems occur with the actual value"""


class ExpectedTypeError(TypeError):
    """Raised when type error problems occur with the expected value"""


class DynamicCallableWithArgsError(TypeError):
    """
    Raised when a dynamic lookup via `attr_is` returns a callable that expects arguments.
    Asserto currently only supports 0-arg callables as part of its dynamic lookup and
    invocations.
    """


class UnsupportedHandlerTypeError(ValueError):
    """Raised when the actual value passed to a handler is not suitable for it to handle."""

    def __init__(self, handler: typing.Type[Handler], method: str, value: typing.Any) -> None:
        super().__init__(f"`{handler.__name__}` cannot accept type: {type(value)} when calling: {method}")



import inspect
import typing


def guard_type_from_method(argument: typing.Any, check_type: typing.Any) -> None:
    """Ensures a particular object has an expected type else raises
    a Type error including the caller of this functions name. This
    should typically only be called by a mixin method performing
    an assertion.

    This unwinds the call stack to derive this information.

    :param argument: The object to check the type of.
    :param check_type: The type it should be of."""
    __tracebackhide__ = True
    current_frame = inspect.currentframe()
    caller = inspect.getouterframes(current_frame, 2)
    method_name = caller[1][3]
    arg_name: str = ""
    if not isinstance(argument, check_type):
        raise TypeError(f"[{method_name}] {arg_name} must be of type {check_type} but was: {type(argument)}.")



import re
import typing

from ._types import EXC_TYPES_ALIAS
from ._types import RE_PATTERN_ALIAS
from ._util import to_iterable
from .descriptors import EnforcedCallable


class ExceptionChecker:
    """
    Encapsulation of the asserto callable exception handling syntax.
    """

    _proxy_val = EnforcedCallable()

    def __init__(
        self,
        exc_types: EXC_TYPES_ALIAS,
        value: typing.Callable[[typing.Any], typing.Any],
        _referent,
        match: typing.Optional[RE_PATTERN_ALIAS] = None,
    ) -> None:
        self.exc_types: typing.Iterable[typing.Type[BaseException]] = to_iterable(exc_types)
        self._proxy_val = value
        self.asserto_ref = _referent
        self.pattern: typing.Optional[re.Pattern[str]] = re.compile(match) if isinstance(match, str) else match

    def when_called_with(self, *args, **kwargs) -> None:
        """
        Call the underlying function with the user supplied arguments;  This returns the result
        of the function back to the asserto instance to enforce error handling & assertion errors
        there.

        If reason is provided; asserto will enforce the exception message is explicitly equal to.

        # Todo: In future support a pattern match.
        # Todo: limitations in this API; what if the called arg has a `reason` attribute?
        """
        try:
            # update 'triggered' status to avoid unnecessary warnings
            self.asserto_ref.triggered = True  # type: ignore[attr-defined]
            self._proxy_val(*args, **kwargs)
            # No exception was raised at all; raise an assertion error.
            self.asserto_ref.error(f"{self._proxy_val} never raised any of: {self.exc_types}")
        except BaseException as e:
            exc_type, exc_string = type(e), str(e)
            if isinstance(e, self.exc_types):  # type: ignore [arg-type]
                if self.pattern:
                    if self.pattern.match(exc_string) is None:
                        # Type matched, but the pattern regex did not, raise.
                        self.asserto_ref.error(
                            f"{exc_type} occurred but did not match pattern: {self.pattern} instead was: {exc_string}"
                        )
                # type matched, no expected pattern, it was a success.
                return
            # The exception was not as expected, raise an assertion error against the type.
            arguments = f"{args, kwargs}" if all((args, kwargs)) else "no arguments"
            self.asserto_ref.error(
                f"{self._proxy_val.__name__} did not raise any of {self.exc_types}.  Instead it raised {exc_type} when called with {arguments}."  # noqa
            )



import typing


class Handler:
    """
    Useful utility functions for handlers.  Subclasses of handlers should raise
    AssertionErrors when things are wrong; the asserto machinery is catching them
    internally to wrap and bubble something more user-friendly back to the user.
    """

    def __init__(self, actual: typing.Any) -> None:
        self.actual = actual

    @staticmethod
    def dispatch_and_raise(fn, expected, error, *args, **kwargs):
        """
        Dispatches a call to an underlying callable and if the return value of the function
        is not equal to expected, raises an AssertionError.
        """
        if fn(*args, **kwargs) != expected:
            raise AssertionError(error)

    def raise_if_length_equals(self, length: int = 0) -> typing.Literal[True]:
        """Checks the length of a Sized implementation and raises a ValueError if it is zero.
        Returns `True` if it succeeds to allow chaining of checks in handler subclasses."""
        if len(self.actual) == length:
            raise ValueError(f"{self.actual} is empty.")
        return True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(actual={self.actual!r})"



import abc
import typing


class Validatable(abc.ABC):
    """
    Base class for custom validation descriptors.
    """

    def __set_name__(self, owner, name):
        self.private_name = "_" + name

    def __get__(self, obj, owner):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abc.abstractmethod
    def validate(self, value: typing.Any) -> None:
        """Raise a ValueError if validation fails."""
        raise NotImplementedError



import re
import typing

EXC_TYPES_ALIAS = typing.Union[typing.Type[BaseException], typing.Iterable[typing.Type[BaseException]]]
CALLABLE_ALIAS = typing.Callable[[typing.Any], typing.Any]
RE_FLAGS_ALIAS = typing.Union[int, re.RegexFlag]
RE_PATTERN_ALIAS = typing.Union[str, typing.Pattern[str]]



""" Error message templates. """
import string
import typing
from dataclasses import dataclass


class CallableTemplate:
    """
    A callable template that delegates to substitute when called.  When called all
    templates will be passed an $actual and $expected value.
    """

    def __init__(self, template: str) -> None:
        self.template = string.Template(template)

    def __call__(self, actual: typing.Optional[typing.Any] = None, expected: typing.Optional[typing.Any] = None):
        return self.template.substitute(actual=actual, expected=expected)


@dataclass(frozen=True)
class StringErrors:
    ends_with: CallableTemplate = CallableTemplate("$actual did not end with $expected")
    starts_with: CallableTemplate = CallableTemplate("$actual did not start with $expected")
    is_alpha: CallableTemplate = CallableTemplate("$actual did not contain only unicode letters")
    is_digit: CallableTemplate = CallableTemplate("$actual did not contain only numeric digits")


@dataclass(frozen=True)
class RegexErrors:
    matches_beginning: CallableTemplate = CallableTemplate("$actual did not begin with $expected")


class Errors:
    """
    A Facade to access the underlying errors via chained API.
    """

    strings = StringErrors
    regex = RegexErrors



import abc
import typing

from ._softly import AssertionErrorContainer


class RaisesErrors(typing.Protocol):
    @abc.abstractmethod
    def check_should_raise(
        self,
        softly: bool,
        cause: typing.Union[AssertionError, str],
        description: typing.Optional[str] = None,
        category: typing.Optional[str] = None,
    ) -> None:
        """Checks and raises immediately for hard failures when not used in a context"""

    @abc.abstractmethod
    def check_soft_raise(self):
        """Checks and raises for soft failures when exiting the context"""

    @abc.abstractmethod
    def reset(self):
        """Reset the soft context failures."""


class ErrorHandler:
    """
    The Error handler.  Responsible for raising the underlying AssertionError's when things
    are amiss for each individual handler instance.
    """

    def __init__(self):
        self.context_failures = AssertionErrorContainer()

    def reset(self) -> None:
        """Clear the underlying sequence of AssertionError instances."""
        self.context_failures.clear()

    def check_should_raise(
        self,
        softly: bool,
        cause: typing.Union[AssertionError, str],
        description: typing.Optional[str] = None,
        category: typing.Optional[str] = None,
    ) -> None:
        """
        Raise the AssertionError or build on the list of soft assertions which
        will be automatically raised when the context exits.  Error can except
        an Assertion error or a string used as message for a newly created one.
        """
        error = cause if isinstance(cause, AssertionError) else self.build_assertion_error(cause, description, category)
        # Check if in 'soft' mode as in the `Asserto` instance was instantiated as a context manager.
        if softly:
            self.context_failures.store(error)
            return
        raise error from None

    def check_soft_raise(self):
        if self.context_failures:
            # The asserto instance had had some soft assertion failures; raise them now.
            raise AssertionError(repr(self.context_failures)) from None

    @staticmethod
    def build_assertion_error(
        cause: str, description: typing.Optional[str] = None, category: typing.Optional[str] = None
    ) -> AssertionError:
        """
        Generates exception messages for a given failure.
        Resolving an assertion error works as follows:

            :: If a description was provided, it is the main exception message
            :: Otherwise `cause` is used, this is typically populated by Asserto
            :: If a category was provided, it is prefixed to the exception message.

        Description gives the user the chance to interject their own bespoke error message on the assertions.

        :param cause: A string for the error message, is always a string at this point.
        :param description: User defined description to provide a custom message on the AssertionError.
        :param category: A prefix for grouping purposes.
        """
        message = description or cause
        return AssertionError(message if category is None else f"[{category}] {message}")



from typing import Any
from typing import Iterable

from typing_extensions import Self as Asserto

from .._decorators import type_check_actual
from .._protocols import Assertable
from .._util import to_iterable

# Todo: end_with offering a `start` and `end`?
# Todo: starts_with offering a `start` and `end`?
# Todo: Docs reuse through __doc__ ?
# Todo: Utility functions and general DRY here
# Todo: This is our first mixin impl, so let's make it perfect before rolling out.


class AssertsStringsMixin(Assertable):
    """Mixin responsible for composing assertions for string types."""

    @type_check_actual((str, Iterable))
    def ends_with(self, suffix: str) -> Asserto:
        """Asserts the actual value ends with a given prefix.  If the actual
        value is an iterable, the last element within it will be compared for
        equality (==) against the suffix.

        :param suffix: The expected substring for the actual value to end with.

        :raises TypeError: When the actual value is not a string or iterable.
        :raises TypeError: When the suffix value is not a string.
        :raises ValueError: When the suffix or actual values are empty.
        :raises ValueError: When the actual value is an empty string.
        :raises AssertionError: When the actual value does not end with the suffix.

        :return: The `Asserto` instance for fluent chaining.
        """
        if isinstance(self.actual, str):
            if not suffix:
                raise ValueError(f"{suffix=} must not be empty.")
            if not isinstance(suffix, str):
                raise TypeError(f"{suffix=} must be a string.")
            if not self.actual.endswith(suffix):
                self.error(f"Expected `{self.actual}` to end with {suffix=} but it did not.")
        else:
            if not self.actual:
                raise ValueError(f"{self.actual} must not be empty.")

            last: Iterable[Any] = to_iterable(self.actual)[-1]
            if last != suffix:
                self.error(f"Expected `{self.actual}` to end with {suffix=} but it did not.")
        return self

    @type_check_actual(str)
    def is_alpha(self) -> Asserto:
        """Asserts the actual value is considered alphabetic.  Empty strings will
        not be considered alphabetic for this case.

        :raises TypeError: If the actual value is not of type string.
        :raises AssertionError: If the actual value is not an alphabetic string.

        :return: The `Asserto` instance for fluent chaining.
        """
        if not self.actual.isalpha():
            self.error(f"{self.actual} is not alphabetic.")

        return self

    @type_check_actual(str)
    def is_digit(self) -> Asserto:
        """Asserts the actual value is a digit string.  Empty strings will not be considered
        digit strings for this case.

        :raises TypeError: If the actual value is not of type string.
        :raises AssertionError: If the actual value is not an alphabetic string.

        :return: The `Asserto` instance for fluent chaining.
        """
        if not self.actual.isdigit():
            self.error(f"{self.actual} is not a digit string.")
        return self

    @type_check_actual(str)
    def is_blank(self) -> Asserto:
        """Asserts the actual value is an empty (blank) string.

        :raises TypeError: If the actual value is not of type string.
        :raises AssertionError: If the actual value is not an empty string
        """
        if len(self.actual):
            self.error(f"{self.actual} was not an empty string.")
        return self

    @type_check_actual((str, Iterable))
    def starts_with(self, prefix: str) -> Asserto:
        """Asserts the actual value starts with the prefix.  If the actual value is
        an iterable the first element is compared for equality (==) against the prefix.


        :param prefix: The value to check the actual value starts with.

        :return: The `Asserto` instance for fluent chaining.

        """
        if not isinstance(prefix, str):
            raise TypeError(f"starts_with prefix must be a string, not: {type(prefix)}")
        if not prefix:
            raise ValueError("starts_with cannot be called with an empty prefix string")

        if isinstance(self.actual, Iterable):
            if isinstance(self.actual, str):
                if not self.actual.startswith(prefix):
                    self.error(f"{self.actual} did not begin with {prefix=}")
            else:
                iterable = iter(self.actual)
                first = next(iterable, None)
                if first is None:
                    raise ValueError(f"cannot check if an empty iterable started with {prefix}")
                if first != prefix:
                    self.error(f"{self.actual} did not start with {prefix}")
        return self



import typing

T = typing.TypeVar("T")


def is_iterable(item: typing.Any) -> bool:
    """Checks if an item is an iterable in a python 3 compliant manner.  This encompasses both
    new iterator protocol aswell as older protocol via __getitem__ etc."""
    try:
        iter(item)
        return True
    except TypeError:
        return False


def to_iterable(item: typing.Any) -> typing.Tuple[T, ...]:
    """
    Converts a tuple from any items that can be iterated over.
    :param item: Anything, if iterable is concerted to tuple otherwise is a single element tuple of T.
    :return: A tuple of items
    """
    try:
        return tuple(iter(item))
    except TypeError:
        return (item,)


def is_namedtuple_like(obj: typing.Any) -> bool:
    """
    Check if an object is likely a named tuple
    :param obj:
    :return:
    """
    t = type(obj)
    bases = t.__bases__
    if len(bases) != 1 and bases[0] != tuple:
        return False
    fields = getattr(obj, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(f) == str for f in fields)



import types
import typing

from ._asserto import Asserto
from ._decorators import update_triggered


def asserto(actual: typing.Any, warn_unused: bool = False) -> Asserto:
    """
    Retrieve an appropriate asserter for the type of value.
    :param actual: The value to compare against later and defer a type specific asserter from.
    :param warn_unused: Emit a warning if not a single assertion was performed to detect user errors.
    :return: An instance of an asserter
    """
    return Asserto(actual, warn_unused)


assert_that = asserto


def register_assert(func: types.FunctionType) -> typing.Callable[[typing.Any], typing.Any]:
    """
    # Todo: There is a lot of edge cases here that I'm not (yet) aware of.
    Automatically registers a user defined callable (function) to all future instances of Asserto.
    This is done magically using setattr and the functions must be imported before instantiating
    instances of Asserto to benefit from them,  store your custom assertion functions somewhere
    loaded early in your code and bind them before instantiating `Asserto` instances.

    Some criteria for user defined assertion functions must be met:
        :: They must use a `self` param first which will be an instance of Asserto
        :: They must be function types with a distinct __name__ to avoid overwriting existing attrs
        :: They must NOT be lambda functions as they have no distinct __name__
        :: They must NOT end in _is as Asserto does some dynamic dispatch using __getattr__ relying on that

    assertable can be called directly and passed your function, or apply it as a decorator to your
    assertion methods.  Examples of both cases are:

        Examples
            :: Usage

            from asserto import assertable

            @register_assert  # Option 1.
            def custom_assert(self):
                if not isinstance(self.actual, MyClass):
                    self.error("my assertion error message")

            register_assert(custom_assert)  # Option 2.
    """
    if not isinstance(func, types.FunctionType):
        raise ValueError("Binding functions must be of function types.")
    name = getattr(func, "__name__")
    # Todo: Can FunctionTypes have no __name__ ?
    if name.endswith("_is"):
        # Due to how asserto handles its dynamic lookups;
        raise ValueError("Binding functions cannot end with `_is`")
    if name == "<lambda>":
        raise ValueError("Binding functions does not support lambdas, they have no name")
    # Register a new function to the class.
    setattr(Asserto, name, update_triggered(func))
    return func



import pprint
import typing


class AssertionErrorContainer:
    """
    An object for storing AssertionError instances when asserto is being used in a soft
    context.  Instances are stored in an underlying sequence and this provides an improved
    API for registering and displaying the errors in a meaningful manner.
    """

    def __init__(self) -> None:
        self.errors: typing.List[AssertionError] = []

    def register_error(self, error: AssertionError) -> None:
        self.errors.append(error)

    def __bool__(self) -> bool:
        """Allow the container to be checked that some errors have occurred."""
        return bool(self.errors)

    def __len__(self) -> int:
        return len(self.errors)

    def clear(self) -> None:
        """Reset all soft failures."""
        self.errors.clear()

    def store(self, error: AssertionError) -> None:
        """Register a new AssertionError."""
        self.errors.append(error)

    def __repr__(self) -> str:
        # Todo: Outline the passes as well as the failures.
        # Todo: Improved tracebacks etc.
        return f"{len(self.errors)} Soft Assertion Failures\n" + pprint.pformat(self.errors, indent=4)



from ._api import assert_that
from ._api import asserto
from ._api import register_assert
from ._asserto import Asserto
from ._exceptions import UnsupportedHandlerTypeError
from ._warnings import NoAssertAttemptedWarning

__all__ = (
    "asserto",
    "register_assert",
    "Asserto",
    "NoAssertAttemptedWarning",
    "UnsupportedHandlerTypeError",
    "assert_that",
)



from typing import Any
from typing import Protocol
from typing import Union


class HasActualValue(Protocol):
    """A simple interface for something containing a `actual` value.
    This helps to support static type hinting with the mixins used
    to compose asserto."""

    @property
    def actual(self) -> Any:
        ...


class CanError(Protocol):
    """A simple interface for something that can raise an AssertionError."""

    def error(self, cause: Union[AssertionError, str]) -> Any:
        ...


class Assertable(HasActualValue, CanError):
    ...



import typing

from ._abc import Validatable


class EnforcedCallable(Validatable):
    """
    A Simple check that an argument is a callable type.
    """

    def validate(self, value: typing.Any) -> None:
        if not callable(value):
            raise ValueError(f"{value} is not callable.")


T = typing.TypeVar("T")


class ValidatesInstanceOf(Validatable, typing.Generic[T]):
    """Check if the value is an instance of multiple types."""

    def __init__(self, *types) -> None:
        self.types = types

    def validate(self, value: typing.Any) -> None:
        if isinstance(value, self.types) is False:
            raise ValueError(f"{value} was not an instance of any of: {self.types}")



from ._base import BaseHandler
from ._handler import Handler
from ._numeric import NumberHandler
from ._regex import RegexHandler

__all__ = ("RegexHandler", "BaseHandler", "NumberHandler", "Handler")



import numbers

from ._base import Handler


class NumberHandler(Handler):
    """A Handler for handling numeric types of data.  Non instances of
    pythons float will be failed instantly by the descriptor,
    so it is safe to assume that all methods here are acting on a stored
    actual value that is in fact, numeric.

    Todo: Better support for datetimes and time deltas etc, these arguably should be comparable
    but should they reside in a separate handler?
    """

    def __init__(self, actual: float) -> None:
        super().__init__(actual)
        self._enforce_is_number()

    def is_zero(self) -> None:
        """Asserts that the value is a numeric type and is equal to 0"""
        if self.actual != 0:
            raise AssertionError(f"Expected {self.actual} to be 0 but it was not.")

    def is_not_zero(self) -> None:
        """Asserts that the value is numeric, and it is greater than other"""
        if self.actual == 0:
            raise AssertionError(f"Expected {self.actual} to not be 0 but it was.")

    def is_greater_than(self, other: float) -> None:
        """Asserts that the value is numeric, and it is lesser than other"""
        if self.actual <= other:
            raise AssertionError(f"Expected {self.actual} to be greater than {other}, but it was not.")

    def is_lesser_than(self, other: float) -> None:
        """Asserts that the value is a numeric type and is lesser than other"""
        if self.actual >= other:
            raise AssertionError(f"Expected {self.actual} to be lesser than {other}, but it was not.")

    def is_positive(self) -> None:
        """Asserts that the value is numeric, and is greater than 0"""
        return self.is_greater_than(0)  # type: ignore[arg-type]

    def is_negative(self) -> None:
        """Asserts that the value is numeric, and is lesser than 0"""
        return self.is_lesser_than(0)  # type: ignore[arg-type]

    def is_between(self, low: float, high: float, inclusive: bool = False):
        """Asserts tht the value is numeric and is between lower and higher.  If inclusive
        is true, value is considered between if it equals either the lower or higher bounds.
        """
        if inclusive:
            if low > self.actual > high:
                raise AssertionError(f"Expected {self.actual} to be inclusively between: ({low}, ..., {high})")
        else:
            if self.actual <= low or self.actual >= high:
                raise AssertionError(f"Expected {self.actual} to be between ({low}, ..., {high})")

    def is_not_between(self, low: float, high: float, inclusive: bool = False):
        """Asserts that the value is numeric and is not between a low and high bounds.  If inclusive
        is true, value is considered between if it equals either the low or high bounds."""
        if inclusive:
            if low < self.actual < high:
                raise AssertionError(f"Expected {self.actual} to not be between ({low}, ..., {high})")
        else:
            if low <= self.actual <= high:
                raise AssertionError(f"Expected {self.actual} to not be between ({low}, ..., {high})")

    def _enforce_is_number(self):
        """
        Enforces that the type of the actual number provided is numeric.  Asserto is handling these
        type errors as part of dispatching to rewrite and raise something more appropriate.
        """
        if isinstance(self.actual, bool):
            raise ValueError
        if not isinstance(self.actual, numbers.Number):
            raise ValueError



import functools
import typing

from ._constants import MixinErrorTemplates
from ._protocols import Assertable


def update_triggered(fn: typing.Callable[[typing.Any], typing.Any]) -> typing.Callable[[typing.Any], typing.Any]:
    """
    Track the triggered state on any asserto calls to ensure no instances were created
    and used without calling any assertable methods.
    :param fn: The asserting function
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs) -> typing.Callable[[typing.Any], typing.Any]:
        instance = args[0]
        instance.triggered = True
        return fn(*args, **kwargs)

    return wrapper


def type_check_actual(
    expected_types: typing.Union[typing.Type[typing.Any], typing.Tuple[typing.Type[typing.Any], ...]]
):
    """Decorator that guards the self.actual type for mixin
    methods to avoid repetitive checks on self.actuals type.

    :param expected_types: An iterable of types to check via isinstance.
    """

    def deco(fn: typing.Callable[[typing.Any, typing.Any], Assertable]):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            instance = args[0]
            actual = instance.actual
            if not isinstance(actual, expected_types):
                method_name = fn.__name__
                raise TypeError(
                    MixinErrorTemplates.ASSERTO_TYPE_ERROR.format(actual, expected_types, method_name, type(actual))
                )
            return fn(*args, **kwargs)

        return wrapper

    return deco



from enum import Enum


# StrEnum does not exist in 3.8
class Methods(str, Enum):
    """
    Encapsulation of method names.  These are used as part of the dynamic dispatch
    to various handlers responsible for individual assertion methods.
    """

    # -- Strings

    ENDS_WITH: str = "ends_with"
    STARTS_WITH: str = "starts_with"
    IS_DIGIT: str = "is_digit"
    IS_ALPHA: str = "is_alpha"
    IS_BETWEEN: str = "is_between"
    IS_NOT_BETWEEN: str = "is_not_between"

    # -- Regular Expressions

    MATCH: str = "match"
    SEARCH: str = "search"
    FULLMATCH: str = "fullmatch"
    FINDALL: str = "findall"

    # -- Base Objects

    IS_TRUE: str = "is_true"
    IS_TRUTHY: str = "is_truthy"
    IS_FALSE: str = "is_false"
    IS_FALSY: str = "is_falsy"
    IS_EQUAL_TO: str = "is_equal_to"
    IS_NOT_EQUAL_TO: str = "is_not_equal_to"
    HAS_LENGTH: str = "has_length"
    IS_INSTANCE: str = "is_instance"
    HAS_SAME_IDENTITY_AS: str = "has_same_identity_as"
    DOES_NOT_HAVE_SAME_IDENTITY_AS: str = "does_not_have_same_identity_as"
    IS_NONE: str = "is_none"
    IS_NOT_NONE: str = "is_not_none"

    # -- Numeric
    IS_ZERO: str = "is_zero"
    IS_NOT_ZERO: str = "is_not_zero"
    IS_GREATER_THAN: str = "is_greater_than"
    IS_LESSER_THAN: str = "is_lesser_than"
    IS_POSITIVE: str = "is_positive"
    IS_NEGATIVE: str = "is_negative"


class MixinErrorTemplates:
    ASSERTO_TYPE_ERROR = "value passed to asserto({0}) must be of type {1} when calling {2} but it was {3}."



from __future__ import annotations

import types
import typing
import warnings

from ._constants import Methods
from ._error_handling import ErrorHandler
from ._error_handling import RaisesErrors
from ._exc_handling import ExceptionChecker
from ._exceptions import DynamicCallableWithArgsError
from ._exceptions import UnsupportedHandlerTypeError
from ._types import EXC_TYPES_ALIAS
from ._types import RE_FLAGS_ALIAS
from ._types import RE_PATTERN_ALIAS
from ._util import is_namedtuple_like
from ._warnings import NoAssertAttemptedWarning
from .handlers import BaseHandler
from .handlers import Handler
from .handlers import NumberHandler
from .handlers import RegexHandler
from .mixins import AssertsStringsMixin

# Todo: base: `tidy up docstrings`
# Todo: base `remove duplication here`
# Todo: Api feels cumbersome with decorators; can we improve DRY-ness?
# Todo: Consider short hand methods; are they worth it vs a single API?


class Asserto(AssertsStringsMixin):
    """
    The entrypoint into asserting objects.

    :param actual: ...
    :param warn_unused: ...
    """

    def __init__(
        self,
        actual: typing.Any,
        warn_unused: bool = False,
        error_handler: typing.Type[RaisesErrors] = ErrorHandler,
    ):
        self._actual = actual
        self._triggered = False
        self.warn_unused = warn_unused
        self._in_context = False
        self._error_handler = error_handler()
        self.category: typing.Optional[str] = None
        self.description: typing.Optional[str] = None

    @property
    def actual(self) -> typing.Any:
        return self._actual

    @actual.setter
    def actual(self, value: typing.Any) -> None:
        self._actual = value

    def error(self, cause: typing.Union[AssertionError, str]) -> Asserto:
        """
        The single point of assertion failing.  All functions delegate here to raise the underlying
        assertion errors.
        :param cause: A reason for the failure. if description was set; it takes precedence.
        :return: The `Asserto` instance for fluency
        """
        self._error_handler.check_should_raise(self._in_context, cause, self.category, self.description)
        return self

    def set_category(self, category: str) -> Asserto:
        """
        Set the category for the assertion.  Categories are prefixed to the assertion
        messages, for example:

            Example:
                Usage::
                    asserto(25).with_category("foo").is_equal_to(26)
                    `AssertionError(["foo"] 25 was not equal to: 26)`
        Args:
            category: The Category to group the assertion under.
        """
        self.category = category
        return self

    def described_as(self, description: str) -> Asserto:
        """
        Set the full `AssertionError`` message to a custom reason.  If this is
        invoked anything after the category (if also set) will be user defined.
        :param description: The reason to display if an `AssertionError` is raised.
        :return: The `Asserto` instance for fluency.
        """
        self.description = description
        return self

    def should_raise(self, exceptions: EXC_TYPES_ALIAS, match: typing.Optional[str] = None) -> ExceptionChecker:
        """
        Wraps the actual value into a callable if it is callable itself;
        :param exceptions: The type of exception expected.
        :param match: An optional message to enforce against when the function is called later.
        :return: The `Asserto` instance for fluency
        """
        return ExceptionChecker(exc_types=exceptions, value=self.actual, _referent=self, match=match)

    # Todo: should_not_raise

    def is_between(self, low: float, high: float, inclusive: bool = False):
        """
        Asserts that the actual value is between a low and high bounds.  If inclusive is true
        the actual value is considered between if it is equal to either of those bounds.
        """
        return self._dispatch(NumberHandler, Methods.IS_BETWEEN, low, high, inclusive)

    def is_between_inclusive(self, low: float, high: float) -> Asserto:
        """
        Asserts that the actual value is equal to or between the log and high bounds.
        """
        return self.is_between(low, high, inclusive=True)

    is_between_incl = is_between_inclusive

    def is_not_between(self, low: float, high: float, inclusive: bool = False):
        """
        Asserts that the actual value is between a **NOT** between a low and high bounds.  If inclusive is true
        the actual value is only considered not between if it is strictly less than the low bounds and higher than
        the high bounds, if it equals either it is considered to be between.
        """
        return self._dispatch(NumberHandler, Methods.IS_NOT_BETWEEN, low, high, inclusive)

    def is_not_between_inclusive(self, low: float, high: float):
        """
        Asserts that the actual value is not between the low and high bounds.  if the actual value
        is equal to the low or high bounds it is considered to be between them
        """
        return self.is_not_between(low, high, inclusive=True)

    is_not_between_incl = is_not_between_inclusive

    def match(self, pattern: RE_PATTERN_ALIAS, flags: RE_FLAGS_ALIAS = 0) -> Asserto:
        """
        Asserts that the actual value provided matches (at least in part) from the beginning of it
        the pattern provided.  This is only a 'begins with' partial match.  Opt for `fullmatch` to
        perform a pattern match on the entirety of the actual value.

        :param pattern: The regular expression pattern to use; r"" is encouraged.
        :param flags: An integer (or RegexFlag) representing flags to apply.
        """
        return self._dispatch(RegexHandler, Methods.MATCH, pattern, flags)

    def search(self, pattern: RE_PATTERN_ALIAS, flags: RE_FLAGS_ALIAS = 0) -> Asserto:
        """
        Asserts that the actual value provided has at least one single match of the pattern
        at some point within it.

        :param pattern: The regular expression pattern to use; r"" is encouraged.
        :param flags: An integer (or RegexFlag) representing flags to apply.
        """
        return self._dispatch(RegexHandler, Methods.SEARCH, pattern, flags)

    def fullmatch(self, pattern: RE_PATTERN_ALIAS, flags: RE_FLAGS_ALIAS = 0) -> Asserto:
        """
        Asserts that the actual value provided wholly matches the pattern provided. Providing
        `^` & `$` is not necessary as they are implicitly inferred.

        :param pattern: The regular expression pattern to use; r"" is encouraged.
        :param flags: An integer (or RegexFlag) representing flags to apply.
        """
        return self._dispatch(RegexHandler, Methods.FULLMATCH, pattern, flags)

    def findall(self, pattern: RE_PATTERN_ALIAS, count: int, flags: RE_FLAGS_ALIAS = 0) -> Asserto:
        """
        Asserts that the total count of non overlapping occurrences of pattern is equal to count.

        :param pattern: The regular expression pattern to use; r"" is encouraged.
        :param count: The expected number of elements expected in the fullmatch returned sequence.
        :param flags: An integer (or RegexFlag) representing flags to apply.
        """
        return self._dispatch(RegexHandler, Methods.FINDALL, pattern, count, flags)

    def is_true(self) -> Asserto:
        """
        Asserts that the actual value is explicitly True.  This uses identity checks internally, to
        check if a value is considered `truthy` use `is_truthy()` instead.
        """
        return self._dispatch(BaseHandler, Methods.IS_TRUE)

    def is_truthy(self) -> Asserto:
        """
        Asserts that the actual value is True in a boolean context.  bool(actual) is called internally
        and the outcome is asserted to be `True`.
        """
        return self._dispatch(BaseHandler, Methods.IS_TRUTHY)

    def is_false(self) -> Asserto:
        """
        Asserts that the actual value is explicitly False.  This uses identity checks internally, to
        check if a value is considered `falsy` use `is_falsy()` instead.
        """
        return self._dispatch(BaseHandler, Methods.IS_FALSE)

    def is_falsy(self) -> Asserto:
        """
        Asserts that the actual value is False in a boolean context.  bool(actual) is called internally
        and the outcome is asserted to be `False`.
        """
        return self._dispatch(BaseHandler, Methods.IS_FALSY)

    def is_equal_to(self, other: typing.Any) -> Asserto:
        """
        Compares the value against `other` for equality.

        :param other: The other object to compare against.
        :return: The instance of `Asserto` to chain asserts.
        """
        return self._dispatch(BaseHandler, Methods.IS_EQUAL_TO, other)

    equals = is_equal_to

    def is_not_equal_to(self, other: typing.Any) -> Asserto:
        """
        Compares the value against `other` for non equality.

        :param other: The other object to compare against.
        :return: The instance of `Asserto` to chain asserts.
        """
        return self._dispatch(BaseHandler, Methods.IS_NOT_EQUAL_TO, other)

    def has_length(self, expected: int) -> Asserto:
        """
        A simple check that the actual value is equal to expected utilising the built in `len(...)`

        :param expected: An int to compare the length against.
        :return: The instance of `Asserto` to chain asserts.
        """
        return self._dispatch(BaseHandler, Methods.HAS_LENGTH, expected)

    def is_instance(self, cls_or_tuple: typing.Union[typing.Any, typing.Iterable[typing.Any]]) -> Asserto:
        """
        Checks if the value provided is either:

            :: A direct subclass.
            :: An indirect subclass.
            :: A virtual subclass registered via the abc.

        :param cls_or_tuple: A single Type, or iterable of types to check the object against.
        """
        return self._dispatch(BaseHandler, Methods.IS_INSTANCE, cls_or_tuple)

    def has_same_identity_as(self, other: typing.Any) -> Asserto:
        """
        Checks that the value refers to the same object in memory as `other`.`
        :param other: The other object to compare identity of.
        :return: The instance of `Asserto` to chain asserts.
        """
        return self._dispatch(BaseHandler, Methods.HAS_SAME_IDENTITY_AS, other)

    def does_not_have_same_identity_as(self, other: typing.Any) -> Asserto:
        """
        Checks that the value does not refer to the same object in memory as `other`.
        :param other: The other object to compare identity of.
        :return: The instance of `Asserto` to chain asserts.
        """
        return self._dispatch(BaseHandler, Methods.DOES_NOT_HAVE_SAME_IDENTITY_AS, other)

    def is_none(self) -> Asserto:
        """
        Checks the actual value is None.  Python `NoneType` is a singleton so `is` checks
        are used
        :return: The `Asserto` instance for fluency.
        """
        return self._dispatch(BaseHandler, Methods.IS_NONE)

    def is_not_none(self) -> Asserto:
        """
        Checks the actual value is not None .  Python `None` is a singleton so `is not` checks are
        used.
        :return: The `Asserto` instance for fluency
        """
        return self._dispatch(BaseHandler, Methods.IS_NOT_NONE)

    def is_zero(self) -> Asserto:
        """
        Checks the actual value is zero.
        :return: The `Asserto` instance for fluency.
        """
        return self._dispatch(NumberHandler, Methods.IS_ZERO)

    def is_not_zero(self) -> Asserto:
        """
        Checks the actual value is not zero.
        :return: The `Asserto` instance for fluency.
        """
        return self._dispatch(NumberHandler, Methods.IS_NOT_ZERO)

    def is_greater_than(self, other: float) -> Asserto:
        """Asserts that the value is numeric, and it is greater than other

        :param other: A number.Number to compare the value against.
        :return: This Asserto instance.
        """
        return self._dispatch(NumberHandler, Methods.IS_GREATER_THAN, other)

    is_more_than = is_greater_than

    def is_lesser_than(self, other: float) -> Asserto:
        """Asserts that the value is numeric, and it is lesser than other

        :param other: A number.Number to compare the value against.
        :return: This Asserto instance.
        """
        return self._dispatch(NumberHandler, Methods.IS_LESSER_THAN, other)

    is_less_than = is_lesser_than

    def is_positive(self) -> Asserto:
        """Asserts that the value is numeric, and is greater than 0

        :return: This Asserto instance.
        """
        return self._dispatch(NumberHandler, Methods.IS_POSITIVE)

    def is_negative(self) -> Asserto:
        """Asserts that the value is numeric, and is lesser than 0

        :return: This Asserto instance.
        """
        return self._dispatch(NumberHandler, Methods.IS_NEGATIVE)

    def _dispatch(self, handler: typing.Type[Handler], method: str, *args, **kwargs) -> Asserto:
        """
        Delegate a check to an underlying handler instance.

        :param error_template: (Optional) The error message to raise on failure.

        Arbitrary args & kwargs to pass through to the handler method.

        Note: Debugging this with an IDE can yield unrealistic as debugging tends to insert
        arbitrary code into the stack and this relies on frame inspection.
        """
        __tracebackhide__ = True  # pytest magic.
        self._triggered = True
        # for now allow this to be bypassed as not all methods have a handler defined.
        try:
            handler_instance = handler(self.actual)  # descriptors enforce types here.
        except ValueError:
            # The type of actual is not viable for the given handler.
            raise UnsupportedHandlerTypeError(handler, method, self.actual) from None
        except KeyError:
            raise KeyError(f"{method} does not have a dedicated handler.") from None
        assertion_method: typing.Optional[types.MethodType] = getattr(handler_instance, method)
        if not callable(assertion_method):
            raise TypeError(f"assertion method was not a bound method on the handler {handler_instance}")
        try:
            _ = assertion_method(*args, **kwargs)
        except AssertionError as e:
            if description := self.description:
                e = AssertionError(description)
            self.error(e)
        # Fall through type & value errors; we don't need to do anything in particular for them, just bubble em up.
        # (for now anyway).
        return self

    @staticmethod
    def _warn_not_triggered() -> None:
        """
        Triggers a warning if an asserto instance was created and no assertion methods was called
        to highlight potential user errors.
        """
        warnings.warn("Asserto instance was created and never used", NoAssertAttemptedWarning, 2)

    def __getattr__(self, item: str) -> typing.Callable[[typing.Any], typing.Any]:
        """
        Adds the capability to object class or instance attributes dynamically.  Supports user defined
        object types as well as built in mapping types.  In the case of a mapping type the attribute
        name will be the key in the dictionary and value called its value to check for equality.

        :param item: The attribute name to lookup
        """
        if not item.endswith("_is"):
            raise AttributeError(f"unknown assertion method: {item}")
        key_attr = item[:-3]
        named_tuple_like = is_namedtuple_like(self.actual)
        mapping_like = isinstance(self.actual, typing.Iterable) and hasattr(self.actual, "__getitem__")
        failure = None

        if not hasattr(self.actual, key_attr):
            # It's not an attribute on the wrapped `actual` value.
            if not named_tuple_like and mapping_like:
                if key_attr not in self.actual:
                    failure = f"{self.actual!r} missing key: {key_attr}"
            else:
                failure = f"{self.actual!r} missing attribute: {key_attr}"

        def _dynamic_callable(*args):
            self._triggered = True  # Dynamic wrapper has been invoked!
            if failure:
                self.error(failure)
            if len(args) != 1:
                raise TypeError(f"Dynamic assertion takes 1 argument but {len(args)} was given. {args}")
            value = self.actual[key_attr] if not named_tuple_like and mapping_like else getattr(self.actual, key_attr)
            if callable(value):
                try:
                    lookup = value()
                except TypeError:
                    raise DynamicCallableWithArgsError(f"{key_attr} expects arguments, this is not supported") from None
            else:
                lookup = value
            expected = args[0]
            if lookup != expected:
                self.error(f"{lookup} was not equal to: {expected}")
            return self

        return _dynamic_callable

    def __repr__(self) -> str:
        return f"Asserto(value={self.actual}, category={self.category})"

    def __enter__(self) -> Asserto:
        """
        Enter a `soft` context mode;  During this context `AssertionErrors` are silently
        ignored until the context is exited at which point any assertions will cause test
        failure(s) and raise the sequence of AssertionError's in the order in which they
        occurred.
        :return: The instance of `Asserto`.
        """
        self._in_context = True
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]] = None,
        exc_val: typing.Optional[BaseException] = None,
        exc_tb: typing.Optional[types.TracebackType] = None,
    ):
        if self.warn_unused and not self._triggered:
            self._warn_not_triggered()
        self._error_handler.check_soft_raise()
        self._error_handler.reset()



import typing

from ._handler import Handler


class BaseHandler(Handler):
    """
    A handler for all objects.
    """

    def __init__(self, actual: typing.Any) -> None:
        super().__init__(actual)

    def is_true(self) -> None:
        """
        Asserts the actual value is explicitly True.
        """
        if not self.actual:
            raise AssertionError(f"{self.actual} was not True")

    def is_truthy(self) -> None:
        """
        Asserts the actual value is True in a boolean context.
        """
        if not bool(self.actual):
            raise AssertionError(f"{self.actual} was not truthy")

    def is_false(self) -> None:
        """
        Asserts the actual value is explicitly False.
        """
        if self.actual:
            raise AssertionError(f"{self.actual} was not False")

    def is_falsy(self) -> None:
        """
        Asserts the actual value is False in a boolean context.
        """
        if bool(self.actual):
            raise AssertionError(f"{self.actual} was not falsy")

    def is_equal_to(self, other: typing.Any) -> None:
        if self.actual != other:
            raise AssertionError(f"{self.actual} is not equal to: {other}")

    def is_not_equal_to(self, other: typing.Any) -> None:
        if self.actual == other:
            raise AssertionError(f"{self.actual} was equal to: {other}")

    def has_length(self, expected: int) -> None:
        if not isinstance(expected, int) or expected < 0:
            raise ValueError(f"{expected} must be an int and greater than 0")

        if len(self.actual) != expected:
            raise AssertionError(f"Length of: {self.actual!r} was not equal to: {expected!r}")

    def is_instance(self, *other: typing.Any) -> None:
        if not isinstance(self.actual, other):
            raise AssertionError(f"{self.actual} was not an instance of: {other}")

    def has_same_identity_as(self, other: typing.Any) -> None:
        if self.actual is not other:
            raise AssertionError(f"{self.actual} does not share identity with: {other}")

    def does_not_have_same_identity_as(self, other: typing.Any) -> None:
        if self.actual is other:
            raise AssertionError(f"{self.actual} shares identity with: {other}")

    def is_none(self) -> None:
        if self.actual is not None:
            raise AssertionError(f"{self.actual} is not None")

    def is_not_none(self) -> None:
        if self.actual is None:
            raise AssertionError(f"{self.actual} is None")
